// ***************************************************************************
// Implementation System Includes
// ***************************************************************************

#include <Bitmap.h>
#include <Screen.h>

#include <posix/stdlib.h>
#include <posix/time.h>

// ***************************************************************************
// Implementation Project Includes
// ***************************************************************************

#include "RotoSaver.h"

// ***************************************************************************
// Implementation Constant Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Type Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Variable Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Declarations
// ***************************************************************************

// ***************************************************************************
// Global Variable Definitions
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions: Lifecycle
// ***************************************************************************

// ===========================================================================
// Constructor
// ===========================================================================

RotoSaver::RotoSaver(BMessage* archive, image_id image)
	: BScreenSaver(archive, image)
{
	// Initialize attributes.
	sourceBitmap = 0;
	destBitmap   = 0;
  
	// Feed random generator seed
	time_t currentTime;
	time(&currentTime);
	srand(currentTime);
}

// ===========================================================================
// Destructor
// ===========================================================================

RotoSaver::~RotoSaver(void) 
{ 
	// Do nothing.
}

// ***************************************************************************
// PUBLIC Member Function Definitions: Operators
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions: Operations
// ***************************************************************************

// ===========================================================================
// StartConfig
// ===========================================================================

void
RotoSaver::StartConfig(BView* view) 
{ 
	BRect rect = view->Bounds();
	BStringView* string;
	
	string = new BStringView(BRect(15, 0, rect.right, 20), 
		B_EMPTY_STRING,
		"Rotosaver 1.2");
	string->SetFont(be_bold_font);
	string->SetFontSize(12);
	view->AddChild(string);
    
    string = new BStringView(
		BRect(10, rect.bottom-65, rect.right, rect.bottom-50), 
		B_EMPTY_STRING, 
		"By Matthijs Hollemans");
	string->SetFont(be_plain_font);
	string->SetFontSize(12); 
	view->AddChild(string);

	string = new BStringView(
		BRect(10, rect.bottom-40, rect.right, rect.bottom-25), 
		B_EMPTY_STRING, 
		"matthijs@weirdexperiments.com");
	string->SetFont(be_plain_font);
	string->SetFontSize(12);
	view->AddChild(string);

	string = new BStringView(
		BRect(10, rect.bottom-25, rect.right, rect.bottom-10), 
		B_EMPTY_STRING, 
		"http://www.weirdexperiments.com");
	string->SetFont(be_plain_font);
	string->SetFontSize(12);
    view->AddChild(string);
} 

// ===========================================================================
// StartSaver
// ===========================================================================
   
status_t
RotoSaver::StartSaver(BView* view, bool preview) 
{ 
	// We don't have a preview.
	if (preview)
	{
		return B_ERROR;
	}
	
	// Get information about the screen.
	BScreen screen(B_MAIN_SCREEN_ID);
	BRect screenFrame = screen.Frame();
	colorSpace = screen.ColorSpace();

	// Figure out whether we can handle the screen's color space.
	if ((B_CMAP8 != colorSpace)
	&&  (B_RGB15 != colorSpace)
	&&  (B_RGB16 != colorSpace)
	&&  (B_RGB32 != colorSpace))
	{
		return B_ERROR;
	}
	
	// Store it in the source bitmap.
	sourceBitmap = new BBitmap(screenFrame, colorSpace);
	if (sourceBitmap != 0)
	{
		screen.ReadBitmap(sourceBitmap, true);
		//_get_screen_bitmap_(sourceBitmap, screenFrame, true);
	}
	else
	{
		return B_ERROR;
	}
  
	// Also create a destination bitmap.
	destBitmap = new BBitmap(screenFrame, colorSpace);
	if (destBitmap == 0)
	{
		return B_ERROR;
	}

	// All set to save the screen!	
	return B_OK; 
} 

// ===========================================================================
// StopSaver
// ===========================================================================

void
RotoSaver::StopSaver(void)
{
	// Get rid of the bitmaps.
	delete sourceBitmap;
	delete destBitmap;
}

// ===========================================================================
// Draw
// ===========================================================================
   
void
RotoSaver::Draw(BView* view, int32 frame) 
{
	// Refresh every couple of seconds.
	if (frame % (4*1000000/50000) != 0)
	{
		return;
	}

	// Figure out dimensions of the bitmaps.
	int32 width = (int32) (sourceBitmap->Bounds().Width() + 1); 
	int32 height = (int32) (sourceBitmap->Bounds().Height() + 1);
	int32 centerX = width/2;
	int32 centerY = height/2;
   
	// Pointers for 8-bit bitmaps.
	int8* source8;   
	int8* dest8;

	// Pointers for 16-bit bitmaps.
	int16* source16;
	int16* dest16;
	
	// Pointers for 32-bit bitmaps.
	int32* source32;
	int32* dest32;
  
	// Get pointers to source and destination bitmaps.
	switch (colorSpace)
	{
		case B_RGB32:
		{
			source32 = (int32*) sourceBitmap->Bits();
			dest32   = (int32*) destBitmap->Bits();
			break;
		}
		case B_RGB15:
			// Fall through.
		case B_RGB16:
		{
			source16 = (int16*) sourceBitmap->Bits();
			dest16   = (int16*) destBitmap->Bits();
			break;
		}
		case B_CMAP8:
		{
			source8  = (int8*) sourceBitmap->Bits();
			dest8    = (int8*) destBitmap->Bits();
			break;
		}
	}

	// Randomly create rotation angle, zoom factor and panning offsets.
	// Note the funny calls to rand(). According to Bjarne Stroustrup, 
	// this is more accurate than simply "rand() % range". Okay, boss.
	float rotationAngle = (double(rand()) / RAND_MAX) * 360;
	float zoomFactor = 0.1 + (double(rand()) / RAND_MAX) * 2;
	float panX = (double(rand()) / RAND_MAX) * width;
	float panY = (double(rand()) / RAND_MAX) * height; 
    
	// Calculate "Right" vector.
	float radAngle = rotationAngle * 6.28318531/360;
	float rightX = cos(radAngle) * zoomFactor;
	float rightY = sin(radAngle) * zoomFactor;

	// Calculate "Down" vector (perpendicular to "right" vector).
	float downX = -rightY;
	float downY = rightX; 
 
	// Calculate source bitmap's start coordinates.
	float I1 = panX - rightX*centerX - downX*centerY;
	float J1 = panY - rightY*centerX - downY*centerY;

	// Save some time by pre-defining some variables 
	// that we use in the loop. Not that it helps.
	float I2, J2;
	int32 I3, J3, offset;
    
	// Loop through the rows of the destination bitmap.
	for (int32 i=0; i<height; ++i)
	{
		I2 = I1;
		J2 = J1;

		// Loop through the columns of the destination bitmap.
		for (int32 j=0; j<width; ++j)
		{
			// Calculate x-position in the source bitmap.
			I3 = (centerX + (int32) I2) % width;
			if (I3 < 0)
			{ 
				I3 += width;
			}
      
			// Calculate y-position in the source bitmap.
			J3 = (centerY + (int32) J2) % height;
			if (J3 < 0)
			{ 
				J3 += height;
			}

			// Calculate pixel offset in the source bitmap.
			offset = width*J3 + I3;

			// Copy pixel from source bitmap to destination bitmap.
			switch (colorSpace)
			{
				case B_RGB32:
				{
					*dest32 = *(source32 + offset);
					dest32++;
					break;
				}
				case B_RGB15:
					// Fall through.
				case B_RGB16:
				{
					*dest16 = *(source16 + offset);
					dest16++;
					break;
				}      
				case B_CMAP8:
				{
					*dest8 = *(source8 + offset);
					dest8++;
					break;
				}            
			}
      
			// Step to the "right" in the source image.
			I2 += rightX;
			J2 += rightY;
		}
    
		// Step "down" in the source image.
		I1 += downX;
		J1 += downY;
	}

	// Use fastest possible drawing mode. Hopefully.
	view->SetDrawingMode(B_OP_COPY);

	// Draw destination bitmap into the view.
	view->DrawBitmap(destBitmap);//, destBitmap->Bounds(), view->Bounds());
}

// ***************************************************************************
// PUBLIC Member Function Definitions: Access
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions: Inquiry
// ***************************************************************************

// ***************************************************************************
// PROTECTED Member Function Definitions
// ***************************************************************************

// ***************************************************************************
// PRIVATE Member Function Definitions
// ***************************************************************************

// ***************************************************************************
// Helper Function Definitions
// ***************************************************************************

// ***************************************************************************
// Global Function Definitions
// ***************************************************************************

// ===========================================================================
// instantiate_screen_saver
// ===========================================================================

BScreenSaver*
instantiate_screen_saver(BMessage* msg, image_id id) 
{ 
	return new RotoSaver(msg, id); 
} 
   
// ***************************************************************************
// Implementation Function Definitions
// ***************************************************************************
